import re
import string

from django.core.exceptions import ValidationError
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _


class RegexValidator:
    regex = NotImplemented
    message = NotImplemented
    code = NotImplemented
    help_text = NotImplemented

    def validate(self, password, user=None):
        if not re.search(self.regex, password):
            raise ValidationError(self.message, code=self.code)

    def get_help_text(self):
        return self.help_text


class DigitValidator(RegexValidator):
    regex = r"[0-9]"
    message = _("This password does not contain any digits (0-9).")
    code = "password_no_digit"
    help_text = _("Your password must contain at least one digit (0-9).")


class UppercaseValidator(RegexValidator):
    regex = r"[A-Z]"
    message = _("This password does not contain any uppercase characters (A-Z).")
    code = "password_no_upper"
    help_text = _("Your password must contain at least one uppercase character (A-Z).")


class LowercaseValidator(RegexValidator):
    regex = r"[a-z]"
    message = _("This password does not contain any lowercase characters (a-z).")
    code = "password_no_lower"
    help_text = _("Your password must contain at least one lowercase character (a-z).")


class SymbolValidator(RegexValidator):
    regex = "|".join(re.escape(c) for c in string.punctuation)
    message = format_lazy(
        _("This password does not contain any special characters ({})."),
        string.punctuation,
    )
    code = "password_no_symbol"
    help_text = format_lazy(
        _("Your password must contain at least one special character ({})."),
        string.punctuation,
    )



from oauth2_provider import views as oauth2_views

from django.contrib.auth.models import AnonymousUser
from django.utils.decorators import method_decorator

from ..csp.decorators import hidp_csp_protection


def _has_prompt_create(request):
    # Check if the prompt=create parameter is present in the request.
    # This parameter is used to explicitly signal that user desires to
    # create a new account rather than authenticate using an existing identity.
    # https://openid.net/specs/openid-connect-prompt-create-1_0.html
    return request.GET.get("prompt") == "create"


@method_decorator(hidp_csp_protection, name="dispatch")
class AuthorizationView(oauth2_views.AuthorizationView):
    registration_url = "hidp_accounts:register"

    def get(self, request, *args, **kwargs):
        if _has_prompt_create(request):
            # Switch request.user to AnonymousUser. This forces handle_no_permission
            # to issue a redirect instead of raising a PermissionDenied exception if
            # a user is currently logged-in.
            self.request.user = AnonymousUser()
            return self.handle_no_permission()
        return super().get(request, *args, **kwargs)

    def get_login_url(self):
        if _has_prompt_create(self.request):
            # The current URL is used as the redirect URL after registration.
            # Drop the prompt=create parameter to return to the authorization flow,
            # without ending up in a redirect loop.
            query = self.request.GET.copy()
            query.pop("prompt")
            self.request.META["QUERY_STRING"] = query.urlencode()
            return self.registration_url
        return super().get_login_url()


@method_decorator(hidp_csp_protection, name="dispatch")
class RPInitiatedLogoutView(oauth2_views.RPInitiatedLogoutView):
    template_name = "hidp/accounts/logout_confirm.html"



import contextlib

from django.http import Http404

from ..federated.providers.base import OIDCClient

_registry = {}


def configure_oidc_clients(*clients):
    """
    Configure OIDC clients for the HIdP application.

    Note:
    This function should be called **once** in the application configuration.
    Subsequent calls will overwrite the registered clients.

    Arguments:
        *clients (OIDCClient):
            One or more OIDCClient instances to register.

    """
    _registry.clear()
    for client in clients:
        if not isinstance(client, OIDCClient):
            raise TypeError(f"Expected OIDCClient, got {type(client).__name__!r}")
        if client.provider_key in _registry:
            raise ValueError(f"Duplicate provider key: {client.provider_key!r}")
        else:
            _registry[client.provider_key] = client


def get_oidc_client(provider_key):
    """
    Retrieve an OIDC client by provider key.

    Arguments:
        provider_key (str):
            The provider key of the client to retrieve.

    Returns:
        OIDCClient:
            The OIDC client instance.

    Raises:
        KeyError:
            If the provider key is not registered.
    """
    try:
        return _registry[provider_key]
    except KeyError:
        raise KeyError(
            f"No OIDC client registered for provider key: {provider_key!r}"
        ) from None


def get_oidc_client_or_none(provider_key):
    """
    Retrieve an OIDC client by provider key or None if provider can not be found.

    Arguments:
        provider_key (str):
            The provider key of the client to retrieve.

    Returns:
        OIDCClient | None:
            The OIDC client instance or None.
    """
    with contextlib.suppress(KeyError):
        return get_oidc_client(provider_key) if provider_key else None
    return None


def get_oidc_client_or_404(provider_key):
    """
    Retrieve an OIDC client by provider key or raise a 404 error if not found.

    Arguments:
        provider_key (str):
            The provider key of the client to retrieve.

    Returns:
        OIDCClient:
            The OIDC client instance.
    """
    with contextlib.suppress(KeyError):
        return get_oidc_client(provider_key)
    raise Http404(f"OIDC Client not found: {provider_key!r}") from None


def get_registered_oidc_clients():
    """
    Retrieve a list of registered OIDC clients.

    Returns:
        list of OIDCClient:
            A list of OIDC client instances.
    """
    return list(_registry.values())



from django.core.management import BaseCommand

from ...email_verification import remove_stale_unverified_accounts


class Command(BaseCommand):
    DEFAULT_MAX_DAYS = 90
    help = (
        "Remove accounts that have not been verified within"
        " a specific number of days after creation"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--days",
            type=int,
            default=self.DEFAULT_MAX_DAYS,
            help=(
                f"Maximum number of days an account is allowed to be unverified:"
                f" defaults to {self.DEFAULT_MAX_DAYS}."
            ),
        )

        parser.add_argument(
            "--dry-run",
            action="store_true",
            help=(
                "Show the number of accounts that would be removed,"
                " without performing the removal."
            ),
        )

    def handle(self, *args, days, dry_run, **options):
        self.stdout.write(
            f"Removing accounts that have not been verified within {days} days..."
        )

        unverified_users_count = remove_stale_unverified_accounts(
            days=days, dry_run=dry_run
        )

        if dry_run:
            self.stdout.write(
                self.style.WARNING(
                    f"{unverified_users_count} unverified account(s) would be removed."
                )
            )
        else:
            self.stdout.write(
                self.style.SUCCESS(
                    f"Successfully removed {unverified_users_count}"
                    f" unverified account(s)."
                )
            )



"""Configuration for Django OAuth Toolkit (DOT) in the HIdP application."""

from datetime import timedelta

OAUTH2_PROVIDER = {
    # The number of seconds an access token remains valid.
    # Requesting a protected resource after this duration will fail.
    # Keep this value high enough so clients can cache the token for
    # a reasonable amount of time.
    #
    # Default in DOT is 1 hour (3600 seconds).
    "ACCESS_TOKEN_EXPIRE_SECONDS": timedelta(hours=12).total_seconds(),
    # The number of seconds an authorization code remains valid.
    # Requesting an access token after this duration will fail.
    # RFC6749 Section 4.1.2 recommends a 10 minute (600 seconds) duration.
    #
    # Default in DOT is 1 minute (60 seconds).
    "AUTHORIZATION_CODE_EXPIRE_SECONDS": timedelta(minutes=10).total_seconds(),
    # The number of seconds before a refresh token gets removed from the database
    # by the cleartokens management command.
    #
    # NOTE: This value is completely ignored when validating refresh tokens.
    #
    # If cleartokens runs daily the maximum delay before a refresh token
    # is removed is REFRESH_TOKEN_EXPIRE_SECONDS + 1 day.
    #
    # Default in DOT is None (never remove refresh tokens).
    "REFRESH_TOKEN_EXPIRE_SECONDS": timedelta(days=90).total_seconds(),
    # The number of seconds between when a refresh token is first used
    # and when it is expired.
    # The most common case for this are native mobile applications
    # that run into issues of network connectivity during the refresh cycle
    # and are unable to complete the full request/response life cycle.
    # Without a grace period the application only has a consumed
    # refresh token and the only recourse is to have the user re-authenticate.
    #
    # Default in DOT is 0 (no grace period).
    "REFRESH_TOKEN_GRACE_PERIOD_SECONDS": timedelta(minutes=10).total_seconds(),
    # Enable OpenID Connect support.
    # Default in DOT is False.
    "OIDC_ENABLED": True,
    # Enable and configure RP-Initiated Logout
    # Default in DOT is False
    "OIDC_RP_INITIATED_LOGOUT_ENABLED": True,
    "OIDC_RSA_PRIVATE_KEY": None,
    # A list of scopes that can be requested by clients, with descriptions.
    "SCOPES": {
        # Default OpenID Connect scope
        # https://openid.net/specs/openid-connect-basic-1_0.html#Scopes
        "openid": "OpenID Connect",
        # OpenID Connect profile scope
        "profile": "View basic profile information",
        # OpenID Connect email scope
        "email": "View email address",
    },
    "DEFAULT_SCOPES": ["openid"],
    # Custom OAuth2Validator that maps OIDC scopes to the correct user attributes
    "OAUTH2_VALIDATOR_CLASS": "hidp.oidc_provider.oauth_validators.OAuth2Validator",
}

# XXX: Everything above this line is included verbatim in the documentation!
#      If this line number changes, update `literalinclude` in the
#      "Adding/overriding Django OAuth Toolkit settings" section of
#      `docs/configure-as-oidc-provider.md` to end at the correct line number.


def get_oauth2_provider_settings(
    *,
    OIDC_RSA_PRIVATE_KEY: str,
):
    """
    Configure Django OAuth Toolkit for the HIdP application.

    Arguments:
        OIDC_RSA_PRIVATE_KEY: str
            The private RSA key used for OpenID Connect (OIDC) support.

            Generate a key using:

                openssl genrsa -out 'oidc.key' 4096

            The contents of the key file should be set as the OIDC_RSA_PRIVATE_KEY.

    Usage:

        In the Django settings module, call this function with the necessary
        configuration values and assign the result to the OAUTH2_PROVIDER setting:

            OAUTH2_PROVIDER = configure(OIDC_RSA_PRIVATE_KEY=...)
    """
    return OAUTH2_PROVIDER | {
        "OIDC_RSA_PRIVATE_KEY": OIDC_RSA_PRIVATE_KEY,
    }



from django_otp.plugins.otp_static.models import StaticDevice

from django.urls import reverse

from hidp.accounts.mailers import BaseMailer
from hidp.otp.devices import get_device_for_user


class BaseOTPUserMailer(BaseMailer):
    def __init__(self, user, *, base_url):
        super().__init__(base_url=base_url)
        self.user = user

    def get_recipients(self):
        return [self.user.email]

    def get_context(self, extra_context=None):
        return super().get_context(
            {
                "otp_management_url": self.base_url
                + reverse("hidp_otp_management:manage")
            }
            | (extra_context or {})
        )


class OTPConfiguredMailer(BaseOTPUserMailer):
    subject_template_name = "hidp/otp/email/configured_subject.txt"
    email_template_name = "hidp/otp/email/configured_body.txt"
    html_email_template_name = "hidp/otp/email/configured_body.html"


class OTPDisabledMailer(BaseOTPUserMailer):
    subject_template_name = "hidp/otp/email/disabled_subject.txt"
    email_template_name = "hidp/otp/email/disabled_body.txt"
    html_email_template_name = "hidp/otp/email/disabled_body.html"


class RecoveryCodeUsedMailer(BaseOTPUserMailer):
    subject_template_name = "hidp/otp/email/recovery_code_used_subject.txt"
    email_template_name = "hidp/otp/email/recovery_code_used_body.txt"
    html_email_template_name = "hidp/otp/email/recovery_code_used_body.html"

    def get_context(self, extra_context=None):
        device = get_device_for_user(self.user, StaticDevice)

        return super().get_context(
            {
                "recovery_codes_count": device.token_set.count(),
            }
            | (extra_context or {})
        )


class RecoveryCodesRegeneratedMailer(BaseOTPUserMailer):
    subject_template_name = "hidp/otp/email/recovery_codes_regenerated_subject.txt"
    email_template_name = "hidp/otp/email/recovery_codes_regenerated_body.txt"
    html_email_template_name = "hidp/otp/email/recovery_codes_regenerated_body.html"



import secrets


def hidp_csp_protection(view_func):
    def _wrapped_view(request, *args, **kwargs):
        request.hidp_csp_nonce = secrets.token_urlsafe(128)
        response = view_func(request, *args, **kwargs)

        if "Content-Security-Policy" not in response.headers:
            response.headers["Content-Security-Policy"] = (
                f"script-src 'nonce-{request.hidp_csp_nonce}' 'strict-dynamic';"
                f" object-src 'none'; base-uri 'none';"
            )
        return response

    return _wrapped_view



from django.test import TestCase, override_settings

from hidp.config import checks

# This module doubles as the ROOT_URLCONF for these tests
urlpatterns = []


# Simulate a poorly configured Django project
@override_settings(
    INSTALLED_APPS=[
        "django.contrib.contenttypes",
        "django.contrib.auth",
        "hidp.accounts",
        "tests.custom_user",
    ],
    AUTH_USER_MODEL="auth.User",
    MIDDLEWARE=[],
    USE_TZ=False,
    OAUTH2_PROVIDER=None,
    ROOT_URLCONF=__name__,  # This module is the ROOT_URLCONF
)
class TestConfigChecks(TestCase):
    def test_required_apps_not_installed(self):
        self.assertEqual(
            checks.check_installed_apps(),
            [
                checks.E001,
            ],
            msg="Expected an error because of missing required apps.",
        )

    def test_required_middlewares_not_installed(self):
        self.assertEqual(
            checks.check_middleware(),
            [
                checks.E002,
            ],
        )

    def test_use_tz_not_set(self):
        self.assertEqual(
            checks.check_use_tz(),
            [
                checks.E003,
            ],
        )

    def test_user_model_not_set_correctly(self):
        self.assertEqual(
            checks.check_user_model(),
            [
                checks.E004,
            ],
        )

    def test_oauth2_provider_not_configured(self):
        self.assertEqual(
            checks.check_oauth2_provider(),
            [
                checks.E005,
            ],
        )

    def test_urls_not_configured(self):
        self.assertEqual(
            checks.check_login_url(),
            [
                checks.E006,
            ],
        )

    def test_oidc_provider_apps_not_installed(self):
        self.assertEqual(
            checks.check_oidc_provider_installed_apps(),
            [
                checks.E008,
            ],
        )

    def test_django_otp_installed_but_hidp_otp_not_installed(self):
        self.assertEqual(
            checks.check_hidp_otp_installed_apps_when_django_otp_installed(),
            [
                checks.W001,
            ],
        )

    def test_otp_required_apps_not_installed_not_triggered(self):
        self.assertEqual(
            checks.check_otp_installed_apps(),
            [],
        )

    @override_settings(INSTALLED_APPS=["hidp.otp"])
    def test_otp_required_apps_not_installed(self):
        self.assertEqual(
            checks.check_otp_installed_apps(),
            [
                checks.E009,
            ],
        )

    def test_otp_required_middleware_not_installed_not_triggered(self):
        self.assertEqual(
            checks.check_otp_middleware(),
            [],
        )

    @override_settings(INSTALLED_APPS=["hidp.otp"])
    def test_otp_required_middleware_not_installed(self):
        self.assertEqual(
            checks.check_otp_middleware(),
            [
                checks.E010,
            ],
        )



from django.db import IntegrityError
from django.test import TestCase

from hidp.config import oidc_clients
from hidp.federated import models
from hidp.test.factories import user_factories

from .test_providers.example import ExampleOIDCClient


class TestOpenIdConnectionModel(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = user_factories.UserFactory()
        cls.connection = models.OpenIdConnection.objects.create(
            user=cls.user,
            provider_key="example",
            issuer_claim="test-issuer",
            subject_claim="test-subject",
        )

    def test_unique_together(self):
        with self.assertRaisesMessage(
            IntegrityError, "duplicate key value violates unique constraint"
        ):
            models.OpenIdConnection.objects.create(
                # Different user, but same OIDC data
                user=user_factories.UserFactory(),
                provider_key="example",
                issuer_claim="test-issuer",
                subject_claim="test-subject",
            )

    def test_str(self):
        oidc_clients.configure_oidc_clients(ExampleOIDCClient(client_id="test"))
        with self.subTest("Registered provider"):
            self.assertEqual(
                str(self.connection),
                "Example (test-subject)",
            )

        oidc_clients.configure_oidc_clients()
        with self.subTest("Unregistered provider"):
            self.assertEqual(
                str(self.connection),
                "Unknown provider: example (test-subject)",
            )

    def test_get_by_provider_and_claims(self):
        with self.assertNumQueries(1):
            connection = models.OpenIdConnection.objects.get_by_provider_and_claims(
                provider_key="example",
                issuer_claim="test-issuer",
                subject_claim="test-subject",
            )
            self.assertEqual(connection, self.connection)
            self.assertEqual(connection.user, self.user)



from datetime import timedelta
from urllib.parse import urlencode

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils import timezone

from . import tokens

UserModel = get_user_model()


def get_email_verification_required_url(user, *, next_url=""):
    url = reverse(
        "hidp_accounts:email_verification_required",
        kwargs={
            "token": tokens.email_verification_request_token_generator.make_token(user),
        },
    )
    if next_url:
        url += f"?{urlencode({'next': next_url})}"
    return url


def get_verify_email_url(user, *, next_url=""):
    url = reverse(
        "hidp_accounts:verify_email",
        kwargs={
            "token": tokens.email_verification_token_generator.make_token(user),
        },
    )
    if next_url:
        url += f"?{urlencode({'next': next_url})}"
    return url


def remove_stale_unverified_accounts(*, days=90, dry_run=False):
    """
    Remove accounts that are not verified within a given number of days after creation.

    Args:
        days (int):
            Maximum number of days an account is allowed to be
            unverified. Defaults to 90.

        dry_run (bool):
            If `True`, returns the number of accounts that would be removed,
            without performing the removal. Defaults to `False`.

    Returns:
        int:
            The number of accounts that are deleted, or would be deleted
            if `dry_run` is `True`
    """
    unverified_users = UserModel.objects.email_unverified().filter(
        date_joined__lt=timezone.now() - timedelta(days=days),
    )
    unverified_users_count = unverified_users.count()

    if not dry_run:
        unverified_users.delete()

    return unverified_users_count



from django.contrib.auth.backends import ModelBackend

from ...config.oidc_clients import get_oidc_client
from ..models import OpenIdConnection


class OIDCModelBackend(ModelBackend):
    def authenticate(
        self,
        request=None,
        provider_key=None,
        issuer_claim=None,
        subject_claim=None,
    ):
        if any(value is None for value in (provider_key, issuer_claim, subject_claim)):
            # None of the required parameters are provided,
            # skip authentication and let another backend handle it.
            return None

        # Find an OpenID connection. This always runs a query regardless whether
        # the provider_key is a registered OIDC provider or not.
        connection = OpenIdConnection.objects.get_by_provider_and_claims(
            provider_key=provider_key,
            issuer_claim=issuer_claim,
            subject_claim=subject_claim,
        )

        try:
            # Check if the provider_key is a registered OIDC provider
            get_oidc_client(provider_key)
        except KeyError:
            return None

        if connection:
            user = connection.user
            if self.user_can_authenticate(user):
                return user
        return None
